/*

clump.c

- use Monte Carlo method to assess significance
of 2 by n contingency table

(C) 1994-2002 Dave Curtis

Permission is hereby granted for the distribution
and modification of this program for any purpose.

Please acknowledge this source and the following
publication. Most up-to-date copies will be kept at
www.mds.qmul.ac.uk/statgen/dcurtis/software.html

Sham PC & Curtis D. 1995. Monte Carlo tests for associations between
disease and alleles at highly polymorphic loci. Ann Hum Genet. 59: 97-105.

This file should compile OK with ANSI C style compilers. The following
command should be approximately correct to compile under Unix.

cc clump.c -o clump -lm

david.curtis@qmul.ac.uk
http://www.mds.qmw.ac.uk/statgen/dcurtis.html
*/

#define VERSION "v2.4"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fisher.h"

#ifndef RAND_MAX
#include <limits.h>
#define RAND_MAX INT_MAX
#endif

#define MAXCOL 500
#define MAXROW 2
#define MAXLINELENGTH (10*MAXCOL)

#define error(s,t) error_func(__LINE__,__FILE__,s,t)

FILE *ftable;

error_func(unsigned l,char *f,char *s1,char *s2)
{
fprintf(stderr,"Error on line %d of source file %s:\n %s%s\n\n",l,f,s1,s2);
exit(1);
return 0;
}

int myrand()
{
static int r[47];
int i,ret;
if (!(r[0]&&r[1]&&r[2]))
  for (i=0;i<47;++i) r[i]=rand();
i=47.0*rand()/((float)RAND_MAX+1.0);
ret=r[i];
r[i]=rand();
return ret;
}

double chistat(double x,double df)
{
if (x==0.0) return 1.0;
else return 1.0-fisher(df,5000,x/df);
}

tab_print(FILE *fo,float *tab,int r,int c)
{
int i,j;
for (i=0;i<r;++i)
  {
  for (j=0;j<c;++j)
    fprintf(fo,"%4.0f ",tab[i*c+j]);
  fprintf(fo,"\n");
  }
}

float get_chisq(float *tab,int r,int c)
{
float mar_r[MAXROW],mar_c[MAXCOL],N,ob,ex,chisq1;
/* don't use exp! (it's a function) */
int i,j;
N=0;
for (i=0;i<r;++i) mar_r[i]=0.0;
for (j=0;j<c;++j) mar_c[j]=0.0;
for (i=0;i<r;++i)
  for (j=0;j<c;++j)
    {
    ob=tab[i*c+j];
    mar_r[i]+=ob;
    mar_c[j]+=ob;
    N+=ob;
    }
if (N==0) return 0.0;
chisq1=0.0;
for (i=0;i<r;++i)
  for (j=0;j<c;++j)
    {
    ob=tab[i*c+j];
    ex=mar_r[i]*mar_c[j]/N;
    if (ex==0.0) continue;
    chisq1+=(ob-ex)*(ob-ex)/ex;
    }
return chisq1;
}

void generate_tab(float *tab,int r,int c,float *prob,float *given_mar_r,
  float *given_mar_c,float N)
{
int i,j;
float mar_r[MAXROW],mar_c[MAXCOL],x,cum_prob;
for (i=0;i<r;++i) mar_r[i]=given_mar_r[i];
for (j=0;j<c;++j) mar_c[j]=given_mar_c[j];
for (i=0;i<r;++i)
  for (j=0;j<c;++j)
    tab[i*c+j]=0;

while (N>0.001)
/* avoid rounding problems for non-integer N */
  {
  x=(myrand()+1.0)/(RAND_MAX+1.0);
  /* avoid x=0, following suggestion from Nathalie Malo, Galileo Genomics */
  for (j=0,cum_prob=0;j<c-1;++j)
    {
    cum_prob+=mar_c[j]/N;
    if (x<=cum_prob) break;
    }
  x=(myrand()+1.0)/(RAND_MAX+1.0);
  for (i=0,cum_prob=0;i<r-1;++i)
    {
    cum_prob+=mar_r[i]/N;
    if (x<=cum_prob) break;
    }
  ++tab[i*c+j];
  --mar_r[i];
  --mar_c[j];
  --N;
  }
return;
}

int clump_tab3(float *tab,int r,int c,float *mar_r,
  float *mar_c,float N,float *max_chisq,int *which_col)
{
/*
compare one column against the rest, except for columns with
small expected values
*/
int i,j,n_cols_OK;
float new_tab[4],chisq;
if (r!=2) { error("Must be two rows in table",""); return 0; }
for (j=0;j<2;++j)
  for (i=0;i<2;++i)
    new_tab[i*2+j]=0.0;
*max_chisq=0.0;
*which_col=-1;
n_cols_OK=0;
for (j=0;j<c;++j)
  {
  if (mar_r[0]*mar_c[j]/N<5 || mar_r[1]*mar_c[j]/N<5) continue;
  ++n_cols_OK;
  for (i=0;i<2;++i)
    {
    new_tab[i*2]=tab[i*c+j];
    new_tab[i*2+1]=mar_r[i]-tab[i*c+j];
    }
  chisq=get_chisq(new_tab,2,2);
  if (chisq>*max_chisq) { *max_chisq=chisq; *which_col=j; }
  }
return n_cols_OK;
}

int old_clump_tab(float *tab,int r,int c,float *mar_r,
  float *mar_c,float N,float *new_tab)
{
/*
collapse together columns with small expected values
*/
int i,j,min_c,next_min_c,new_c;
float min_exp,next_min_exp;
if (r!=2) { error("Must be two rows in table",""); return 0; }
for (j=0;j<c;++j)
  for (i=0;i<2;++i)
    new_tab[i*c+j]=tab[i*c+j];
min_c=next_min_c=-1; /* just to avoid warnings */
new_c=c;
for(;;)
  {
  min_exp=next_min_exp=99999;
  for (j=0;j<new_c;++j)
    for (i=0;i<2;++i)
      {
      if (mar_c[j]*mar_r[i]<min_exp)
        {
        min_exp=mar_c[j]*mar_r[i];
        min_c=j;
        }
      }
  for (j=0;j<new_c;++j)
    for (i=0;i<2;++i)
      {
      if (mar_c[j]*mar_r[i]<next_min_exp
           && j!=min_c)
        {
        next_min_exp=mar_c[j]*mar_r[i];
        next_min_c=j;
        }
      }
  if (min_exp/N<5)
    {
    if (new_c<3) return 0; /* down to 2x2 table + still few expected */
    for (i=0;i<2;++i)
      new_tab[i*new_c+next_min_c]+=new_tab[i*new_c+min_c];
    mar_c[next_min_c]+=mar_c[min_c];
    for (j=min_c;j<new_c-1;++j)
      mar_c[j]=mar_c[j+1];
    for (i=1;i>=0;--i)
      for (j=i*new_c+min_c;j<2*new_c-1;++j)
        new_tab[j]=new_tab[j+1];
        /* shift second row first, then first one */
    --new_c;
    }
  else break;
  }
return new_c;
}

void swap(float *swapped,float *clumped,int which,float *tab,int c,
  int *in_col,int *n_in_col)
{
int i,j;
for (i=0;i<2;++i)
 for (j=0;j<2;++j)
   swapped[i*2+j]=clumped[i*2+j];
if (in_col[which]) /* go from 1 to 0 */
  {
  swapped[1]-=tab[which];
  swapped[0]+=tab[which];
  swapped[3]-=tab[c+which];
  swapped[2]+=tab[c+which];
  }
else
  {
  swapped[0]-=tab[which];
  swapped[1]+=tab[which];
  swapped[2]-=tab[c+which];
  swapped[3]+=tab[c+which];
  }
}

int clump_tab2(float *tab,int r,int c,float *mar_r,
  float *mar_c,float N,float *new_tab,int *which_ones)
{
/*
clump columns to produce 2x2 table with maximum chi-square
*/
int i,j,in_col[MAXCOL],n_in_col[2],which,maxwhich;
float swapped[4],chisq1,maxchisq1;
if (r!=2) { error("Must be two rows in table",""); return 0; }
for (i=0;i<2;++i)
 for (j=0;j<2;++j)
   new_tab[i*2+j]=0.0;
n_in_col[0]=n_in_col[1]=0;
for (j=0;j<c;++j)
  if (tab[j]/mar_c[j]<mar_r[0]/N)
    {
    in_col[j]=0;
    ++n_in_col[0];
    new_tab[0]+=tab[j];
    new_tab[2]+=tab[c+j];
    }
  else
    {
    in_col[j]=1;
    ++n_in_col[1];
    new_tab[1]+=tab[j];
    new_tab[3]+=tab[c+j];
    }
if (n_in_col[0]==0 || n_in_col[1]==0)
  swap(swapped,new_tab,0,tab,c,in_col,n_in_col);
  /* make sure at least one in each column */
for (;;)
  {
  maxchisq1=chisq1=get_chisq(new_tab,2,2);
  for (i=0;i<c;++i)
    {
    if (n_in_col[0]==1 && in_col[i]==0 ||
         n_in_col[1]==1 && in_col[i]==1)
         continue; /* last one in column */
    swap(swapped,new_tab,i,tab,c,in_col,n_in_col);
    if (get_chisq(swapped,2,2)>maxchisq1)
      {
      maxchisq1=get_chisq(swapped,2,2);
      maxwhich=i;
      }
    }
  if (maxchisq1==chisq1) break; /* can't do better */
  swap(swapped,new_tab,maxwhich,tab,c,in_col,n_in_col);
  if (in_col[maxwhich]==0)
    {
    in_col[maxwhich]=1;
    --n_in_col[0];
    ++n_in_col[1];
    }
  else
    {
    in_col[maxwhich]=0;
    --n_in_col[1];
    ++n_in_col[0];
    }
  for (i=0;i<2;++i)
   for (j=0;j<2;++j)
     new_tab[i*2+j]=swapped[i*2+j];
  }
for (i=0,j=0;i<c;++i)
  if (in_col[i]) which_ones[j++]=i;
which_ones[j]=-1;
return 1;
}

int do_sim(float *tab,int r,int c,float *chisq0,float *chisq1,
  float *chisq2,float *chisq3,long nsims,
  float *sim_vals0,int *over0,float *sim_vals1,int *over1,
  float *sim_vals2,int *over2,float *sim_vals3,int *over3,int *n_cols_OK,
  int *which_ones,int *which_colmax)
{
float mar_r[MAXROW],mar_c[MAXCOL],old_mar_c[MAXCOL],
   old_clumped[MAXCOL*2],N,ob,clumped[4],prob[MAXCOL*MAXROW];
int i,j,old_c,dummies[MAXCOL],dummy;
long count;
N=0;
old_c=c;
for (i=0;i<r;++i) mar_r[i]=0.0;
for (j=0;j<c;++j) mar_c[j]=0.0;
for (i=0;i<r;++i)
  for (j=0;j<c;++j)
    {
    ob=tab[i*c+j];
    mar_r[i]+=ob;
    mar_c[j]+=ob;
    N+=ob;
    }
*chisq0=get_chisq(tab,r,c);
if (ftable)
  {
  tab_print(ftable,tab,r,c);
  fprintf(ftable,"Normal chisq (T1) = %f\n",*chisq0);
  }
old_c=c;
for (j=0;j<c;++j)
  {
  old_mar_c[j]=mar_c[j];
  for (i=0;i<r;++i)
    old_clumped[i*c+j]=tab[i*c+j];
  }
if ((old_c=old_clump_tab(tab,r,c,mar_r,old_mar_c,N,old_clumped))==0)
/* not a fatal error */
  {
  old_c=2;
  *chisq1=0;
  }
else
  {
  if (ftable) tab_print(ftable,old_clumped,2,old_c);
  *chisq1=get_chisq(old_clumped,2,old_c);
  }
*n_cols_OK=clump_tab3(tab,r,c,mar_r,mar_c,N,chisq3,which_colmax);
if (ftable)
  {
  fprintf(ftable,"Chisq after columns with small expected counts clumped together (T2) = %f\n",*chisq1);
  fprintf(ftable,"Chisq comparing each column against the rest (T3) = %f (out of possible %d cols, using column %d)\n",
    *chisq3,*n_cols_OK,*which_colmax+1);
  }
if (clump_tab2(tab,r,c,mar_r,mar_c,N,clumped,which_ones))
  {
  if (ftable)
    {
    tab_print(ftable,clumped,2,2);
    fprintf(ftable,"Table clumped to maximise chisq\n");
    fprintf(ftable,"Achieved by clumping together columns:\n");
    for (i=0;which_ones[i]>-1;++i) fprintf(ftable,"%d ",which_ones[i]+1);
    fprintf(ftable,"\n");
    }
  *chisq2=get_chisq(clumped,2,2);
  }
else *chisq2=0;
if (ftable)
  {
  fprintf(ftable,"Maximise 2x2 chisq (T4) = %f\n",*chisq2);
  fprintf(ftable,"\n\nBeginning simulations...\n\n");
  }
*over0=*over1=*over2=*over3=0;
for (count=0;count<nsims;++count)
  {
  generate_tab(tab,r,c,prob,mar_r,mar_c,N);
  if ((*sim_vals0=get_chisq(tab,r,c)) >= *chisq0) ++*over0;
  generate_tab(old_clumped,r,old_c,prob,mar_r,old_mar_c,N);
  if ((*sim_vals1=get_chisq(old_clumped,2,old_c)) >= *chisq1) ++*over1;
  if (ftable)
    {
    tab_print(ftable,tab,r,c);
    fprintf(ftable,"Normal chisq (T1) = %f\n",*sim_vals0);
    tab_print(ftable,old_clumped,2,old_c);
    fprintf(ftable,"Chisq of table simulated from collapsed table (T2) = %f\n",*sim_vals1);
    }
  if (clump_tab2(tab,r,c,mar_r,mar_c,N,clumped,dummies))
    {
    if (ftable) tab_print(ftable,clumped,2,2);
    if ((*sim_vals2=get_chisq(clumped,2,2)) >= *chisq2) ++*over2;
    }
  else *sim_vals2=0;
  clump_tab3(tab,r,c,mar_r,mar_c,N,sim_vals3,&dummy);
  if (*sim_vals3>= *chisq3) ++*over3;
  if (ftable)
    {
    fprintf(ftable,"Maximised 2x2 chisq2 (T4) = %f\n",*sim_vals2);
    fprintf(ftable,"Maximum chisq using one column against the rest (T3) = %f\n",*sim_vals3);
    }
  }
return 1;
}

void usage()
{
printf("\n\nUsage: clump infile outfile [tablefile]\n");
}

main(int argc,char *argv[])
{
int c,i,over0,over1,over2,over3,seed,n_cols_OK,tot;
float tab[MAXCOL*2],chisq0,chisq1,chisq2,chisq3,
sim_vals0,sim_vals1,sim_vals2,sim_vals3;
int which_ones[MAXCOL],which_colmax;
long nsims;
char line[MAXLINELENGTH],rest[MAXLINELENGTH];
FILE *fi,*fo;
printf("\nCLUMP %s\n",VERSION);
printf("\nIf you use this program, please cite:\n\n"
"Sham PC, Curtis D. Monte Carlo tests for associations between disease and \n"
"alleles at highly polymorphic loci. Annals of Human Genetics 1995 59 97-105.\n\n");
if (argc<3)
  {
  usage();
  return 1;
  }
fi=fopen(argv[1],"r");
if (fi==NULL) { error("Could not open file: ",argv[1]); return 1; }
fo=fopen(argv[2],"w");
if (fo==NULL) { error("Could not open file: ",argv[2]); return 1; }
if (argc>3)
  {
  ftable=fopen(argv[3],"w");
  if (ftable==NULL) { error("Could not open file: ",argv[3]); return 1; }
  }
fprintf(fo,"Input:\n\n");
if (argc>1) fprintf(fo,"Input file = %s\n",argv[1]);
printf("Reading %s\n",argv[1]);
fscanf(fi," %d ",&c);
printf("Reading data for %d alleles\n",c);
if (!fgets(line,MAXLINELENGTH,fi)) { error("Could not read data",""); return 1; }
for (i=0,tot=0;i<c;++i)
  {
  *rest='\0';
  if (sscanf(line," %f %[^\n]",&tab[i],rest)<1)
    { error("Bad data:",line); return 1; }
  tot+=tab[i];
  strcpy(line,rest);
  }
for (i=0;i<c;++i) fprintf(fo,"%3.0f ",tab[i]);
fprintf(fo,"  First row total = %d\n",tot);
if (!fgets(line,MAXLINELENGTH,fi)) { error("Could not read data",""); return 1; }
for (i=0,tot=0;i<c;++i)
  {
  *rest='\0';
  if (sscanf(line," %f %[^\n]",&tab[c+i],rest)<1)
    { error("Bad data:",line); return 1; }
  tot+=tab[c+i];
  strcpy(line,rest);
  }
for (i=0;i<c;++i) fprintf(fo,"%3.0f ",tab[c+i]);
fprintf(fo,"  Second row total = %d  \n",tot);
fscanf(fi," %ld",&nsims);
printf("Will perform %ld simulations\n",nsims);
printf("Enter random number seed: ");
fscanf(fi," %d",&seed);
fprintf(fo,"Number of simulations = %ld\nRandom number seed = %d\n",nsims,seed);
srand(seed);
printf("\n");
if (fi!=stdin) while (fgets(line,MAXLINELENGTH,fi)) fprintf(fo,"%s",line); /* comments*/
fprintf(fo,"\n\nOutput:\n");

printf("Running simulations...\n");
if (do_sim(tab,2,c,&chisq0,&chisq1,&chisq2,&chisq3,nsims,&sim_vals0,&over0,
     &sim_vals1,&over1,&sim_vals2,&over2,&sim_vals3,&over3,&n_cols_OK,which_ones,&which_colmax)==-1)
  { error("Problem doing simulations",""); return 1; }
printf("...Done.\n");
fprintf(fo,"\nNormal chi-square (T1) was %f\nThis was reached %d times in %ld simulations (p=%f)\n",
  chisq0,over0,nsims,(over0+1.0)/(nsims+1.0));
fprintf(fo,"(p value from chi-squared distribution, %d df, p=%f)\n\n",c-1,chistat(chisq0,(float)(c-1)));
fprintf(fo,"Chi-square from table after collapsing columns with small \nexpected values together (T2) was %f\n\
This was reached %d times in %ld simulations based on the collapsed table. (p=%f)\n",
  chisq1,over1,nsims,(over1+1.0)/(nsims+1.0));
fprintf(fo,"(p value from chi-squared distribution, %d df, p=%f)\n\n",n_cols_OK-1,chistat(chisq1,(float)(n_cols_OK-1)));
fprintf(fo,"Chi-square from comparing each column against the rest (T3) was %f\n\
This was reached %d times in %ld simulations (p=%f)\n",chisq3,over3,nsims,(over3+1.0)/(nsims+1.0));
fprintf(fo,"(number of columns with expected values of 5 or more = %d, column used=%d)\n",
  n_cols_OK,which_colmax+1);
fprintf(fo,"(Uncorrected p value from chi-squared distribution, 1 df, p=%f)\n\n",chistat(chisq3,1.0));
fprintf(fo,"Chi-square from 2x2 table clumped to produce maximum (T4) was %f\n\
This was reached %d times in %ld simulations (p=%f)\n\n",chisq2,over2,nsims,(over2+1.0)/(nsims+1.0));
fprintf(fo,"Achieved by clumping together columns:\n");
for (i=0;which_ones[i]>-1;++i) fprintf(fo,"%d ",which_ones[i]+1);
fprintf(fo,"\n");
fclose(fo);
if (ftable) fclose(ftable);
return 0;
}
